#include "cdce913.h"

#include "Wire.h"
#include <math.h>

#define CDCE913_I2C_ADDRESS 0x65
#define u8 unsigned char
#define i8 char
#define u16 unsigned int
#define i16 int
#define u32 unsigned long
#define i32 long

#define FREQ_125MHZ (125000000 / 1000000)
#define FREQ_150MHZ (150000000 / 1000000)
#define FREQ_175MHZ (175000000 / 1000000)

void CDCE913::writeByte(u8 addr, u8 value)
{
    Wire.beginTransmission(CDCE913_I2C_ADDRESS); // transmit to device CDCE913_I2C_ADDRESS
    Wire.write(/*Address*/ addr & 0x7F | /* 0 - Block; 1 - Byte */ 1 << 7);
    Wire.write(value);
    Wire.endTransmission();    // stop transmitting
}

u8 CDCE913::readByte(u8 addr)
{
    // request 1 bytes from slave device CDCE913_I2C_ADDRESS
    Wire.requestFrom(CDCE913_I2C_ADDRESS, 1, /*Address*/ addr & 0x7F | /* 0 - Block; 1 - Byte */ 1 << 7, 1, true);
    while (Wire.available()) { // slave may send less than requested
        return Wire.read(); // receive a byte as character
    }
}

u8 CDCE913::calc_p(u16 n, u16 m)
{
    u8 result;
    result = 4 - (int)floor(log((double)n / (double)m) * M_LOG2E);
    if (result < 0)
        result = 0;
    return result;
}

CDCE913::CDCE913()
{
    Wire.begin();
//    Serial.begin(9600);
}


bool CDCE913::init(double fIn)
{
    this->fIn = fIn;

    u8 reg = readByte(0x00);
    if ((reg & 0x1) == 0) { // Error
        return false;
    }

    u16 pDiv = 1;//4 * 25;// / 4;

    writeByte(0x01, 0x5);
    writeByte(0x02, 0xB4 | ((pDiv >> 8) & 0x3));
    writeByte(0x03, pDiv);
    writeByte(0x13, 0x01);
    writeByte(0x14, 0x69);//0X6D);

    return true;
}

u32 pow_(u8 p)
{
    u32 res = 1;
    for (int i = 0; i < p; ++i)
        res *= 2;
    return res;
}

bool CDCE913::setFout(double fOut)
{
    double br, fvco_err;
    double min_fvco_err = 300.0;
    int min_n, min_m, min_p, min_q, min_r, min_pDiv, min_fvco;
    int n, m, p, q, r;
    i32 nstroke;
    bool error = true;
    u8 vco_range;
    double fvco;
    u32 pp;

    u16 pDiv;

    for (pDiv = 1; pDiv < 1023; ++pDiv) {
        fvco = fOut * pDiv;
        if (fvco < 80.0 || fvco > 230.0) {
            continue;
        }

        br = fvco / fIn;

        for (m = 511; m > 0; m--) {
            n = floor((double)m * br);

            if (n > 4095 || n < m) {
                continue;
            }

            p = calc_p(n, m);

            if (p > 4) {
                continue;
            }

//            pp = pow(2, p);
            pp = pow_(p);
            nstroke = n * pp;

            q = (int)floor((double)nstroke / (double)m);

            if (q < 16 || q > 63) {
                continue;
            }

            r = nstroke - m * q;

            if (r < 0 || r > 51) {
                continue;
            }

            fvco_err = fabs((fIn * ((double)n / (double)m)) - fvco);

            if (fvco_err < min_fvco_err) {
                min_fvco_err = fvco_err;
                min_n = n;
                min_m = m;
                min_p = p;
                min_q = q;
                min_r = r;
                min_pDiv = pDiv;
                min_fvco = fvco;
                error = false;
            }
        }
    }

    if (min_fvco < FREQ_125MHZ) vco_range = 0;
    else    if ((min_fvco >= FREQ_125MHZ) && (min_fvco < FREQ_150MHZ)) vco_range = 1;
            else    if ((min_fvco >= FREQ_150MHZ) && (min_fvco < FREQ_175MHZ)) vco_range = 2;
                    else vco_range = 3;

    if (error) return false;

    u8 reg18 = min_n >> 4;//(N >> ) & 0xFFF;
    u8 reg19 = ((min_n & 0xF) << 4) | ((min_r >> 5) & 0xF);
    u8 reg1A = ((min_r & 0x1F) << 3) | ((min_q >> 3) & 0x7);
    u8 reg1B = ((min_q & 0x7) << 5) | ((min_p & 0x7) << 2) | (vco_range & 0x03);

    writeByte(0x02, 0xB4 | ((min_pDiv >> 8) & 0x3));
    writeByte(0x03, min_pDiv);
    writeByte(0x18, reg18); // N
    writeByte(0x19, reg19); // NR
    writeByte(0x1A, reg1A); // RQ
    writeByte(0x1B, reg1B); // QP RANGE

//    Serial.print("n=");
//    Serial.print(min_n);
//    Serial.print(" m=");
//    Serial.print(min_m);
//    Serial.print(" p=");
//    Serial.print(min_p);
//    Serial.print(" q=");
//    Serial.print(min_q);
//    Serial.print(" r=");
//    Serial.print(min_r);
//    Serial.print(" vco_range=");
//    Serial.print(vco_range);
//    Serial.print(" pDiv=");
//    Serial.println(min_pDiv);

//    u8 reg;
//    for (int i = 0; i < 0x1F; ++i) {
//        reg = readByte(i);
//        Serial.print(i);
//        Serial.print("=");
//        Serial.print(reg, HEX);
//        if (i % 8 == 0) Serial.println();
//        else Serial.print(" ");
//    }

    return true;
}

